#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "qproxy.h"
#include "client.h"
#include "funcs.h"
#include "io.h"

extern struct buffer_t *cl_in, *sv_in, *cl_out, *sv_out;
extern struct buffer_t *sv_pending, *cl_pending;
extern struct info_t info;
extern struct player_t players[32];
extern int myid;
extern char *rot13(char *);
extern char *models[256], *sound[256];
extern struct entity_t entities[0x200];
int client_unparsed, beepent = 0;
long cli_seqnum, cli_seqack;

struct opcode_t client_packet[] =
{
	{0x03, "Movement", cli_movement},
	{0x04, "Command", cli_command},
	{0x05, "Frame", cli_frame},
	{0x06, "Position", cli_position}
};
void client_parse(void)
{
	char *special;
	int client_port, opcode, x, match;

	cli_seqnum = get_long(cl_in);
	if (cli_seqnum == 0xffffffff) {
		special = get_string(cl_in);
		printf("cli: %s\n", special);
		put_long(cl_out, cli_seqnum);
		put_string(cl_out, special);
		return;
	}
	cli_seqack = get_long(cl_in);
	client_port = get_short(cl_in);

	put_long(cl_out, cli_seqnum);
	put_long(cl_out, cli_seqack);
	put_short(cl_out, client_port);

	while (cl_in->pos < cl_in->len) {
		opcode = get_byte(cl_in);

		match = 0;
		for (x = 0; x < sizeof(client_packet) / sizeof(client_packet[1]); x++) {
			if (opcode == client_packet[x].opcode) {
				fprintf(stderr, "[Client] Opcode %02x (%s)\n", opcode, client_packet[x].desc);
				client_packet[x].handle(opcode);
				match = 1;
			}
		}
		if (!match) {
			fprintf(stderr, "UNPARSED BYTE: %02x\n", opcode);
			clprint(3, "UNPARSED CLIENT BYTE: %02x\n", opcode);
			fflush(stderr);
			client_unparsed++;
			/* try to keep the connection alive */
			put_byte(cl_out, opcode);
		}
	}

	if (cl_in->pos != cl_in->len)
		fprintf(stderr, "<<<<<%d/%d bytes parsed>>>>>\n", cl_in->pos, cl_in->len);

}

void cli_movement(int opcode)
{
	struct movement_t delta[3];
	int x, unknown;

	info.frame = get_byte(cl_in);
	unknown = get_byte(cl_in);

	for (x = 0; x < 3; x++) {
		delta[x].mask = get_byte(cl_in);
		delta[x].angle[0] = (delta[x].mask & 0x01) ? get_angle16(cl_in) : 0;
		delta[x].angle[1] = (delta[x].mask & 0x80) ? get_angle16(cl_in) : 0;
		delta[x].angle[2] = (delta[x].mask & 0x02) ? get_angle16(cl_in) : 0;
		delta[x].speed[0] = (delta[x].mask & 0x04) ? get_short(cl_in) : 0;
		delta[x].speed[1] = (delta[x].mask & 0x08) ? get_short(cl_in) : 0;
		delta[x].speed[2] = (delta[x].mask & 0x10) ? get_short(cl_in) : 0;
		delta[x].flags = (delta[x].mask & 0x20) ? get_byte(cl_in) : 0;
		delta[x].impulse = (delta[x].mask & 0x40) ? get_byte(cl_in) : 0;
		delta[x].load = get_byte(cl_in);
	}
	players[myid].cangle[0] = delta[0].angle[0];
	players[myid].cangle[1] = delta[0].angle[1];
	players[myid].cangle[2] = delta[0].angle[2];
	players[myid].cgoing[0] = delta[0].speed[0];
	players[myid].cgoing[1] = delta[0].speed[1];
	players[myid].cgoing[2] = delta[0].speed[2];
	players[myid].flags = delta[0].flags;
	players[myid].impulse = delta[0].impulse;
	players[myid].load = delta[0].load;
	info.framerate = 1000.0 / players[myid].load;
	
	put_byte(cl_out, opcode);
	put_byte(cl_out, info.frame);
	put_byte(cl_out, unknown);
	
	for (x = 0; x < 3; x++) {
		put_byte(cl_out, delta[x].mask);
		if (delta[x].mask & 0x01) put_angle16(cl_out, delta[x].angle[0]);
		if (delta[x].mask & 0x80) put_angle16(cl_out, delta[x].angle[1]);
		if (delta[x].mask & 0x02) put_angle16(cl_out, delta[x].angle[2]);
		if (delta[x].mask & 0x04) put_short(cl_out, delta[x].speed[0]);
		if (delta[x].mask & 0x08) put_short(cl_out, delta[x].speed[1]);
		if (delta[x].mask & 0x10) put_short(cl_out, delta[x].speed[2]);
		if (delta[x].mask & 0x20) put_byte(cl_out, delta[x].flags);
		if (delta[x].mask & 0x40) put_byte(cl_out, delta[x].impulse);
		put_byte(cl_out, delta[x].load);
	}
}

void cli_command(int opcode)
{
	char *command;

	command = get_string(cl_in);
//	fprintf(stderr, "Client command: %s\n", command);
	printf("client command: %s\n", command);
	
	if (!strncmp(command, "say_team p ", 11) || !strncmp(command, "say_team \"p ", 12)) {
		got_command(command);
	} else {
		put_byte(cl_out, opcode);
		put_string(cl_out, command);
	}
}

void cli_frame(int opcode)
{
	players[myid].frame = get_byte(cl_in);
	fprintf(stderr, "Frame: %02x\n", players[myid].frame);

	put_byte(cl_out, opcode);
	put_byte(cl_out, players[myid].frame);
}

void cli_position(int opcode)
{
	int x;
	float origin[3];
	
	for (x = 0; x < 3; x++)
		origin[x] = get_coord(cl_in);

	put_byte(cl_out, opcode);
	for (x = 0; x < 3; x++)
		put_coord(cl_out, origin[x]);
}
