#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <sys/time.h>

#include "qproxy.h"
#include "io.h"
#include "funcs.h"

extern struct buffer_t *cl_pending, *sv_pending;
struct timer_ll *timers = NULL;
extern char *sounds[256], *models[256];
extern char *lightstyle[64];
extern struct entity_t entities[0x200];
extern struct player_t players[32];
extern int myid, autoaim;
int stufftext = 1;
int is_bright = 0;
float MULT = 1.0;

void make_sound(int idx, int entity, int chan, float vol, float attenuation, float origin[3])
{
	int x;
	long mix = 0;
	
	mix = (entity & 0x3ff) << 3;
	if (vol != 1.0) mix |= 0x8000;
	if (attenuation != 1.0) mix |= 0x4000;
	mix |= chan & 0x07;
	
	put_byte(sv_pending, 0x06);
	put_short(sv_pending, mix);
	if (mix & 0x8000) put_byte(sv_pending, vol * 255.0);
	if (mix & 0x4000) put_byte(sv_pending, attenuation * 64.0);
	put_byte(sv_pending, idx);
	for (x = 0; x < 3; x++)
		put_coord(sv_pending, origin[x]);
	fprintf(stderr, "MAKE Sound %d[%s]: o[%.1f %.1f %.1f] c[%d] v[%f] a[%f] e[%d]\n", idx, sounds[idx], origin[0], origin[1], origin[2], chan, vol, attenuation, entity);
}

int find_model_idx(char *name)
{
	int x;
	
	for (x = 1; models[x]; x++) {
		if (!strcmp(name, models[x]))
			return x;
	}
	return 0;
}

int find_sound_idx(char *name)
{
	int x;
	
	for (x = 1; sounds[x]; x++) {
		if (!strcmp(name, sounds[x]))
			return x;
	}
	return 0;
}

float distance(float pos1[3], float pos2[3])
{
	float x, y, z;
	
	x = pos1[0] - pos2[0];
	y = pos1[1] - pos2[1];
	z = pos1[2] - pos2[2];
	
	return sqrt(x*x + y*y + z*z);
}

void add_timer(long sec, long usec, void *handle)
{
	struct timer_ll *t;
	struct timeval tv;
	
	gettimeofday(&tv, NULL);
	
	t = malloc(sizeof(struct timer_ll));
	
	t->sec = tv.tv_sec + sec;
	t->setsec = sec;
	t->usec = tv.tv_usec + usec;
	t->setusec = usec;
	t->handle = handle;
	t->next = NULL;
	
	if (t->usec > 1000000) {
		t->sec++;
		t->usec -= 1000000;
	}
	
	if (timers)
		t->next = timers;
	timers = t;
}

struct timer_ll *find_timer(long sec, long usec, void *handle)
{
	struct timer_ll *t;
	
	t = timers;
	while (t) {
		if (t->setsec == sec && t->setusec == usec && t->handle == handle)
			return t;
		t = t->next;
	}
	return NULL;
}
void del_timer(struct timer_ll *victim)
{
	struct timer_ll *t, *prev = NULL;
	
	t = timers;
	while (t) {
		if (t == victim)
			break;
		prev = t;
		t = t->next;
	}
	if (!t)
		return;
	if (prev)
		prev->next = t->next;
	else
		timers = t->next;
	free(t);
}

void check_timers(void)
{
	struct timer_ll *t, *next;
	struct timeval tv;
	
	gettimeofday(&tv, NULL);

	t = timers;
	while (t) {
		next = t->next;
		if (t->sec < tv.tv_sec || (t->sec == tv.tv_sec && t->usec <= tv.tv_usec)) {
			t->handle();
			del_timer(t);
		}
		t = next;
	}
}

void beep_quad(void)
{
	int x, stop = 0;
	
	printf("beep quad\n");
	for (x = 0; x < 0x1ff; x++) {
		if (models[entities[x].modelidx]) {
			if (!strcmp(models[entities[x].modelidx], "progs/quaddama.mdl")) {
				printf("found one, making noise\n");
				make_sound(find_sound_idx("weapons/tink1.wav"), x, 3, 1.0, 1.0, entities[x].origin);
				if (distance(players[myid].origin, entities[x].origin) < 300.0) {
					printf("close, stopping\n");
					fflush(stdout);
					stop = 1;
				}
			}
		}
	}
	printf("stop=%d, !stop=%d\n", stop, !stop);
	if (!stop)
		add_timer(1, 0, beep_quad);
}

void got_command(char *text)
{
	text = strchr(text, 'p');
	/* ugly */
	text++;
	text++;
	if (text[strlen(text) - 1] == '"')
		text[strlen(text) - 1] = '\0';

	if (!strncasecmp(text, "rot", 3)) {
		char buf[2048];
		put_byte(cl_pending, 0x04);
		sprintf(buf, "say hAs%s", rot13(text + 4));
		put_string(cl_pending, buf);
	}
	if (!strncasecmp(text, "autoaim", 7)) {
		autoaim = !autoaim;
		clprint(2, "autoaim is now %s\n", autoaim ? "on" : "off");
	}
	if (!strncasecmp(text, "fullbright", 10)) {
		int x;
		
		if (is_bright) {
			for (x = 0; x < 64; x++) {
				put_byte(sv_pending, 0x0c);
				put_byte(sv_pending, x);
				put_string(sv_pending, lightstyle[x]);
			}
		} else {
			for (x = 0; x < 64; x++) {
				put_byte(sv_pending, 0x0c);
				put_byte(sv_pending, x);
				put_string(sv_pending, "z");
			}
		}
	}
	if (!strncasecmp(text, "quad", 4)) {
		printf("adding quad timer\n");
		add_timer(1, 0, beep_quad);
	}
	if (!strncasecmp(text, "mult", 4)) {
		MULT = atof(text + 5);
		clprint(2, "multiplier is now %.1f\n", MULT);
	}
	if (!strncasecmp(text, "nostuff", 7)) {
		stufftext = !stufftext;
	}
}

char *rot13(char *input)
{
	static char temp[2048];
	register char *p = NULL;
	int rotate = 13;

	strncpy(temp, input, sizeof(temp) - 1);
	temp[2047] = 0;
	
	for (p = temp; *p; p++) {
		if (*p >= 'A' && *p <= 'Z')
			*p = (*p - 'A' + rotate) % 26 + 'A';
		else if (*p >= 'a' && *p <= 'z')
			*p = (*p - 'a' + rotate) % 26 + 'a';
	}
	return temp;
}

float get_yaw(float x1, float y1, float x2, float y2)
{
	double adj, opp;
	double angle;
	
	opp = y2 - y1;
	adj = x2 - x1;

	angle = atan(fabs(opp/adj));

	if (adj < 0 && opp >= 0)
		angle = M_PI - angle;
	if (adj < 0 && opp < 0)
		angle = M_PI + angle;
	if (adj >= 0 && opp < 0)
		angle = (2*M_PI) - angle;

	return (float)angle;
}

float get_tilt(float x1, float y1, float z1, float x2, float y2, float z2)
{
	float adj, opp;
	float angle;
	
	opp = sqrt(pow(x2 - x1, 2.0) + pow(y2 - y1, 2.0));
	adj = z2 - z1;

	angle = opp < adj ? M_PI_2 - atan(fabs(opp/adj)) : atan(fabs(adj/opp));

	if (adj > 0)
		angle = -angle;

	return angle;
}

int find_nearest_player(void)
{
	int x, alive_model, player = myid;
	float lowest = 99999.9, dist;
	
	for (x = 0; x < 32; x++) {
		if (!players[x].visible)
			continue;
		if ((!players[x].team || !players[myid].team || strcmp(players[x].team, players[myid].team)) && players[x].frame != 102) {
			if ((dist = distance(players[myid].origin, players[x].origin)) < lowest) {
				lowest = dist;
				player = x;
			}
		}
	}
	return player;
}
