#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "qproxy.h"
#include "io.h"

#ifdef DEBUG
int verbose;
#endif

struct info_t info;
struct player_t players[32];
struct entity_t entities[0x200];
extern int client_unparsed, server_unparsed;
time_t start;

int main(int argc, char *argv[])
{
	int opt, port_listen = DEFAULT_PORT, port_server = DEFAULT_PORT;
	char *ptr;
	
	while ((opt = getopt(argc, argv, "l:Vvh")) != EOF) {
		switch (opt) {
			case 'V':
				fprintf(stderr, "qproxy %s -- |Rain|\nexe: %s %s\n", VERSION, __DATE__, __TIME__);
				return 0;
			case 'l':
				port_listen = atoi(optarg);
				break;
#ifdef DEBUG
			case 'v':
				verbose++;
				break;
#endif
			default:
				show_help(argv[0]);
				return (opt != 'h');
				
		}
	}
	if (optind != argc - 1) {
		show_help(argv[0]);
		return 1;
	}
	if ((ptr = strchr(argv[optind], ':'))) {
		*ptr = 0;
		port_server = atoi(ptr + 1);
	}

	signal(SIGINT, got_sig);
	start = time(NULL);
	
	qw_connect(argv[optind], port_server, port_listen);
	mainloop();
	qw_disconnect();
	
	return 0;
}

void show_help(char *argv0)
{
	fprintf(stderr, "Usage: %s [options] <host[:port]>\n", argv0);
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "\t-V:\tShow version and exit\n");
	fprintf(stderr, "\t-l:\tChange listening port\n");
	fprintf(stderr, "\t-h:\tShow this help\n");
#ifdef DEBUG
	fprintf(stderr, "\t-v:\tIncrease verbosity\n");
#endif
}

void got_sig(int num)
{
	fprintf(stderr, "signal %d recieved\n", num);
	fprintf(stderr, "unparsed packets:\n");
	fprintf(stderr, "client: %d\tserver: %d\n", client_unparsed, server_unparsed);
	fprintf(stderr, "uptime: %ld seconds\n", time(NULL) - start);
	qw_disconnect();
	exit(0);
}
