#include <math.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "qproxy.h"
#include "server.h"
#include "funcs.h"
#include "io.h"

extern struct buffer_t *cl_in, *sv_in, *cl_out, *sv_out;
extern struct buffer_t *sv_pending, *cl_pending;
extern struct info_t info;
extern struct player_t players[32];
extern struct entity_t entities[0x200];
int server_unparsed = 0;
char *sounds[256], *models[256], *centerprint = NULL;
char *lightstyle[64];
long stats[15];
int myid = 0, autoaim = 0;
long ser_seqnum, ser_seqack, stufftext;
extern float MULT;

struct opcode_t server_packet[] =
{
	{0x01, "NOP", ser_nop},
	{0x02, "Disconnect", ser_disconnect},
	{0x03, "Updatestat", ser_updatestat},
	{0x06, "Sound", ser_sound},
	{0x08, "Print", ser_print},
	{0x09, "StuffText", ser_stufftext},
	{0x0a, "SetAngle", ser_setangle},
	{0x0b, "ServerData", ser_serverdata},
	{0x0c, "Lightstyle", ser_lightstyle},
	{0x0e, "UpdateFrags", ser_updatefrags},
	{0x10, "StopSound", ser_stopsound},
	{0x13, "Damage", ser_damage},
	{0x14, "SpawnStatic", ser_spawnstatic},
	{0x16, "SpawnDynamic", ser_spawndynamic},
	{0x17, "TempEnt", ser_tempent},
	{0x18, "Pause", ser_pause},
	{0x1a, "CenterPrint", ser_centerprint},
	{0x1b, "KilledMonster", ser_killedmonster},
	{0x1c, "FoundSecret", ser_foundsecret},
	{0x1d, "StaticSound", ser_staticsound},
	{0x1e, "Intermission", ser_intermission},
	{0x1f, "Finale", ser_finale},
	{0x20, "CDTrack", ser_cdtrack},
	{0x21, "Help", ser_help},
	{0x22, "SmallKick", ser_smallkick},
	{0x23, "BigKick", ser_bigkick},
	{0x24, "UpdatePing", ser_updateping},
	{0x25, "UpdateEnter", ser_updateenter},
	{0x26, "UpdateStatLong", ser_updatestatlong},
	{0x27, "MuzzleFlash", ser_muzzleflash},
	{0x28, "UpdateUserInfo", ser_updateuserinfo},
	{0x29, "Download", ser_download},
	{0x2a, "PlayerInfo", ser_playerinfo},
	{0x2b, "Nails", ser_nails},
	{0x2c, "Choke", ser_choke},
	{0x2d, "PrecacheModels", ser_precachemodels},
	{0x2e, "PrecacheSounds", ser_precachesounds},
	{0x2f, "PacketEntities", ser_entities},
	{0x30, "DeltaPacketEntities", ser_deltaentities},
	{0x31, "MaxSpeed", ser_maxspeed},
	{0x32, "EntGravity", ser_entgravity},
	{0x33, "SetInfo", ser_setinfo},
	{0x34, "ServerInfo", ser_serverinfo},
	{0x35, "UpdatePL", ser_updatepl}
};

void server_parse(void)
{
	char *special;
	int opcode, x, match;

	ser_seqnum = get_long(sv_in);
	if (ser_seqnum == 0xffffffff) {
		special = get_string(sv_in);
		printf("ser: %s\n", special);
		put_long(sv_out, ser_seqnum);
		put_string(sv_out, special);
		return;
	}
	ser_seqack = get_long(sv_in);

	put_long(sv_out, ser_seqnum);
	put_long(sv_out, ser_seqack);

	while (sv_in->pos < sv_in->len) {
		opcode = get_byte(sv_in);

		match = 0;
		for (x = 0; x < sizeof(server_packet) / sizeof(server_packet[1]); x++) {
			if (opcode == server_packet[x].opcode) {
				fprintf(stderr, "[Server] Opcode %02x (%s)\n", opcode, server_packet[x].desc);
				server_packet[x].handle(opcode);
				match = 1;
			}
		}
		if (!match) {
			fprintf(stderr, "UNPARSED BYTE: %02x\n", opcode);
			fflush(stderr);
			clprint(3, "UNPARSED BYTE: %02x\n", opcode);
			server_unparsed++;

			/* try to keep the connection alive */
			put_byte(sv_out, opcode);
		}
	}
	if (sv_in->pos != sv_in->len)
		fprintf(stderr, "<<<<<%d/%d bytes parsed>>>>>\n", sv_in->pos, sv_in->len);
}

void clprint(int level, char *format, ...)
{
	char buf[2048];
	va_list args;
	
	va_start(args, format);
	if (level == 5)
		put_byte(sv_pending, 0x1a);
	else {
		put_byte(sv_pending, 0x08);
		put_byte(sv_pending, level);
	}
	vsprintf(buf, format, args);
	va_end(args);
	
	put_string(sv_pending, buf);
}

void ser_nop(int opcode)
{
	fprintf(stderr, "Odd.  NOP recieved\n");

	put_byte(sv_out, opcode);
}

void ser_disconnect(int opcode)
{
	put_byte(sv_out, opcode);
}

void ser_updatestat(int opcode)
{
	int id, value;
	
	id = get_byte(sv_in);
	value = get_byte(sv_in);

	fprintf(stderr, "stats %d is now %d\n", id, value);
//	clprint(2, "stats %d is now %d (%02x)\n", id, value, value);
	stats[id] = value;

	/*
	 0: health
	 1: 
	 2: weapon model idx
	 3: current weapon ammo
	 4:
	 5:
	 6: shells
	 7: nails
	 8: rockets
	 9: cells
	 10: current weapon [long?]
	 11:
	 12:
	 13:
	 14:
	 15: item flags [long]
	*/
	put_byte(sv_out, opcode);
	put_byte(sv_out, id);
	put_byte(sv_out, value);
}

void ser_sound(int opcode)
{
	long mix;
	float vol, attenuation;
	int idx, chan, x, entity;
	float origin[3];
	
	mix = get_short(sv_in);
	vol = (mix & 0x8000) ? (float)(get_byte(sv_in) / 255.0) : 1.0;
	attenuation = (mix & 0x4000) ? (float)(get_byte(sv_in) / 64.0) : 1.0;
	chan = mix & 0x07;
	entity = (mix >> 3) & 0x3ff;
	idx = get_byte(sv_in);
	for (x = 0; x < 3; x++)
		origin[x] = get_coord(sv_in);
	fprintf(stderr, "Sound %d[%s]: o[%.1f %.1f %.1f] c[%d] v[%f] a[%f] e[%d]\n", idx, sounds[idx], origin[0], origin[1], origin[2], chan, vol, attenuation, entity);
//	clprint(1, "sound: %s@%.1f %.1f %.1f [ent:%d]\n", sounds[idx], origin[0], origin[1], origin[2], entity);
	put_byte(sv_out, opcode);
	put_short(sv_out, mix);
	if (mix & 0x8000) put_byte(sv_out, vol * 255.0);
	if (mix & 0x4000) put_byte(sv_out, attenuation * 64.0);
	put_byte(sv_out, idx);
	for (x = 0; x < 3; x++)
		put_coord(sv_out, origin[x]);
}

void ser_print(int opcode)
{
	int level;
	char *string, *ptr;
	
	level = get_byte(sv_in);
	string = get_string(sv_in);
	
	fprintf(stderr, "Text: [%d]%s\n", level, string);
	
	put_byte(sv_out, opcode);
	put_byte(sv_out, level);
	put_string(sv_out, string);

	if ((ptr = strchr(string, ':'))) {
		if (!strncmp(ptr, ": hAs", 4)) {
			fprintf(stderr, "Encoded message: %s", rot13(string + 5));
			clprint(3, "encoded message: %s", rot13(ptr + 5));
		}
	}
}

void ser_stufftext(int opcode)
{
	char *command;
	
	command = get_string(sv_in);
	
	fprintf(stderr, "stufftext: %s\n", command);
	
	if (stufftext) {
		put_byte(sv_out, opcode);
		put_string(sv_out, command);
	}
}

void ser_setangle(int opcode)
{
	int x;
	float angle[3];
	
	for (x = 0; x < 3; x++)
		angle[x] = get_angle(sv_in);

	fprintf(stderr, "setangle: %.1f %.1f %.1f\n", angle[0], angle[1], angle[2]);
	
	put_byte(sv_out, opcode);
	for (x = 0; x < 3; x++)
		put_angle(sv_out, angle[x]);
}

void ser_serverdata(int opcode)
{
	long version, uptime;
	float var[10];
	char *mapname, *dir, buf[128];
	int x;
	
	version = get_long(sv_in);
	uptime = get_long(sv_in);
	dir = strdup(get_string(sv_in));
	myid = get_byte(sv_in);
	mapname = get_string(sv_in);
	
	for (x = 0; x < 10; x++)
		var[x] = get_float(sv_in);

	memset(&entities, 0, sizeof(entities));
	memset(&players, 0, sizeof(players));
	
	stufftext = 1;

	fprintf(stderr, "version: %ld uptime: %ld id: %d\n", version, uptime, myid);
	fprintf(stderr, "directory: %s\n", dir);
	fprintf(stderr, "map: %s\n", mapname);

	/*
		0: sv_gravity
		1: sv_stopspeed
		2: sv_maxspeed
		3: sv_spectatormaxspeed
		4: sv_accelerate
		5: sv_airaccelerate
		6: sv_wateraccelerate
		7: sv_friction
		8: sv_waterfriction
		9: ? entgravity ?
	*/

	for (x = 0; x < 10; x++) {
		fprintf(stderr, "var[%d] = %f\n", x, var[x]);
//		clprint(2, "var[%d] = %f\n", x, var[x]);
	}

	sprintf(buf, "%s (id: %d, dir: %s)", mapname, myid, dir);

	put_byte(sv_out, opcode);
	put_long(sv_out, version);
	put_long(sv_out, uptime);
	put_string(sv_out, dir);
	put_byte(sv_out, myid);
/*	put_string(sv_out, mapname); */
	put_string(sv_out, buf);
	
	for (x = 0; x < 10; x++)
		put_float(sv_out, var[x]);

	if (myid & 0x80)
		fprintf(stderr, "**spectator mode**\n");
	myid &= ~0x80;

	free(dir);
}

void ser_lightstyle(int opcode)
{
	int num;
	char *style;
	
	num = get_byte(sv_in);
	style = get_string(sv_in);
	
	fprintf(stderr, "lightstyle %d = %s\n", num, style);
	if (lightstyle[num])
		free(lightstyle[num]);
	lightstyle[num] = strdup(style);
	
	put_byte(sv_out, opcode);
	put_byte(sv_out, num);
	put_string(sv_out, style);
}

void ser_updatefrags(int opcode)
{
	int player, frags;
	
	player = get_byte(sv_in);
	frags = get_short(sv_in);
	
	fprintf(stderr, "player %d has %d frags\n", player, frags);
	
	put_byte(sv_out, opcode);
	put_byte(sv_out, player);
	put_short(sv_out, frags);
}

void ser_cdtrack(int opcode)
{
	int track;
	
	track = get_byte(sv_in);
	
	put_byte(sv_out, opcode);
	put_byte(sv_out, track);
}

void ser_stopsound(int opcode)
{
	short mix, channel, entity;
	
	mix = get_short(sv_in);
	channel = mix & 0x07;
	entity = (mix >> 3) & 0x03ff;
	
	put_byte(sv_out, opcode);
	put_short(sv_out, mix);
}

void ser_damage(int opcode)
{
	int armor, health, x;
	float origin[3];
	
	armor = get_byte(sv_in);
	health = get_byte(sv_in);
	for (x = 0; x < 3; x++)
		origin[x] = get_coord(sv_in);
	
	fprintf(stderr, "damage from '%.1f %.1f %.1f': %d health, %d armor\n", origin[0], origin[1], origin[2], health, armor);

	put_byte(sv_out, opcode);
	put_byte(sv_out, armor);
	put_byte(sv_out, health);
	for (x = 0; x < 3; x++)
		put_coord(sv_out, origin[x]);
}

void ser_spawnstatic(int opcode)
{
	int modelidx, frame, colormap, skin, x;
	float origin[3], angle[3];
	
	modelidx = get_byte(sv_in);
	frame = get_byte(sv_in);
	colormap = get_byte(sv_in);
	skin = get_byte(sv_in);
	
	for (x = 0; x < 3; x++) {
		origin[x] = get_coord(sv_in);
		angle[x] = get_angle(sv_in);
	}

	fprintf(stderr, "static ent: %d[%s] f[%d] c[%d] s[%d] o[%.1f %.1f %.1f] a[%.1f %.1f %.1f]\n", modelidx, models[modelidx], frame, colormap, skin, origin[0], origin[1], origin[2], angle[0], angle[1], angle[2]);

	put_byte(sv_out, opcode);
	put_byte(sv_out, modelidx);
	put_byte(sv_out, frame);
	put_byte(sv_out, colormap);
	put_byte(sv_out, skin);
	
	for (x = 0; x < 3; x++) {
		put_coord(sv_out, origin[x]);
		put_angle(sv_out, angle[x]);
	}
}

void ser_spawndynamic(int opcode)
{
	int entnum, x;
	
	entnum = get_short(sv_in);
	entities[entnum].modelidx = get_byte(sv_in);
	entities[entnum].frame = get_byte(sv_in);
	entities[entnum].colormap = get_byte(sv_in);
	entities[entnum].skin = get_byte(sv_in);
	entities[entnum].visible = 0;
	
	for (x = 0; x < 3; x++) {
		entities[entnum].origin[x] = get_coord(sv_in);
		entities[entnum].angle[x] = get_angle(sv_in);
	}

	fprintf(stderr, "dynamic ent: e[%d] %d[%s] f[%d] c[%d] s[%d] o[%.1f %.1f %.1f] a[%.1f %.1f %.1f]\n", entnum, entities[entnum].modelidx, models[entities[entnum].modelidx], entities[entnum].frame, entities[entnum].colormap, entities[entnum].skin, entities[entnum].origin[0], entities[entnum].origin[1], entities[entnum].origin[2], entities[entnum].angle[0], entities[entnum].angle[1], entities[entnum].angle[2]);

	put_byte(sv_out, opcode);
	put_short(sv_out, entnum);
	put_byte(sv_out, entities[entnum].modelidx);
	put_byte(sv_out, entities[entnum].frame);
	put_byte(sv_out, entities[entnum].colormap);
	put_byte(sv_out, entities[entnum].skin);
	
	for (x = 0; x < 3; x++) {
		put_coord(sv_out, entities[entnum].origin[x]);
		put_angle(sv_out, entities[entnum].angle[x]);
	}
}

void ser_pause(int opcode)
{
	int action;
	
	action = get_byte(sv_in);
	
	fprintf(stderr, "game is now %spaused\n", action ? "" : "un");
	
	put_byte(sv_out, opcode);
	put_byte(sv_out, action);
}

void ser_tempent(int opcode)
{
	int type, entity, num, x;
	float origin[3], dest[3];
	
	type = get_byte(sv_in);
	
	switch (type) {
		case 0:
		case 1:
		case 3:
		case 4:
		case 7:
		case 8:
		case 10:
		case 11:
		case 13:
			for (x = 0; x < 3; x++)
				origin[x] = get_coord(sv_in);
			break;
		case 5:
		case 6:
		case 9:
			entity = get_short(sv_in);
			for (x = 0; x < 3; x++)
				origin[x] = get_coord(sv_in);
			for (x = 0; x < 3; x++)
				dest[x] = get_coord(sv_in);
			break;
		case 2:
		case 12:
			num = get_byte(sv_in);
			for (x = 0; x < 3; x++)
				origin[x] = get_coord(sv_in);
			break;
		default:
			fprintf(stderr, "invalid tempent: %d\n", type);
			break;
	}

	put_byte(sv_out, opcode);
	put_byte(sv_out, type);
	
	switch (type) {
		case 0:
		case 1:
		case 3:
		case 4:
		case 7:
		case 8:
		case 10:
		case 11:
		case 13:
			for (x = 0; x < 3; x++)
				put_coord(sv_out, origin[x]);
			break;
		case 5:
		case 6:
		case 9:
			put_short(sv_out, entity);
			for (x = 0; x < 3; x++)
				put_coord(sv_out, origin[x]);
			for (x = 0; x < 3; x++)
				put_coord(sv_out, dest[x]);
			break;
		case 2:
		case 12:
			put_byte(sv_out, num);
			for (x = 0; x < 3; x++)
				put_coord(sv_out, origin[x]);
			break;
	}
}

void ser_centerprint(int opcode)
{
	struct timer_ll *t;
	char *text;
	
	text = get_string(sv_in);

	fprintf(stderr, "centerprint: %s\n", text); 
	
	if (centerprint)
		free(centerprint);
	centerprint = strdup(text);
	
	if ((t = find_timer(2, 0, clear_centerprint)))
		del_timer(t);
	add_timer(2, 0, clear_centerprint);
		
	
	put_byte(sv_out, opcode);
	put_string(sv_out, text);
}

void clear_centerprint(void)
{
	free(centerprint);
	centerprint = NULL;

	add_timer(2, 0, clear_centerprint);
}

void ser_killedmonster(int opcode)
{
	fprintf(stderr, "a monster died\n");
	
	put_byte(sv_out, opcode);
}

void ser_foundsecret(int opcode)
{
	fprintf(stderr, "a secret was found\n");
	
	put_byte(sv_out, opcode);
}

void ser_staticsound(int opcode)
{
	int idx, x;
	float vol, attenuation, origin[3];
	
	for (x = 0; x < 3; x++)
		origin[x] = get_coord(sv_in);
	idx = get_byte(sv_in);
	vol = get_byte(sv_in) / 255.0;
	attenuation = get_byte(sv_in) / 64.0;
	
	fprintf(stderr, "static sound: %d[%s] o[%.1f %.1f %.1f] v[%.2f] a[%.2f]\n", idx, sounds[idx], origin[0], origin[1], origin[2], vol, attenuation);

	put_byte(sv_out, opcode);
	for (x = 0; x < 3; x++)
		put_coord(sv_out, origin[x]);
	put_byte(sv_out, idx);
	put_byte(sv_out, vol * 255.0);
	put_byte(sv_out, attenuation * 64.0);
}

void ser_intermission(int opcode)
{
	float origin[3], angle[3];
	int x;
	
	for (x = 0; x < 3; x++)
		origin[x] = get_coord(sv_in);
	for (x = 0; x < 3; x++)
		angle[x] = get_angle(sv_in);
	
	fprintf(stderr, "intermission: o[%.1f %.1f %.1f] a[%.1f %.1f %.1f]\n", origin[0], origin[1], origin[2], angle[0], angle[1], angle[2]);

	put_byte(sv_out, opcode);
	for (x = 0; x < 3; x++)
		put_coord(sv_out, origin[x]);
	for (x = 0; x < 3; x++)
		put_angle(sv_out, angle[x]);
}

void ser_finale(int opcode)
{
	char *text;
	
	text = get_string(sv_in);
	fprintf(stderr, "finale: %s\n", text);

	put_byte(sv_out, opcode);
	put_string(sv_out, text);
}

void ser_help(int opcode)
{
	fprintf(stderr, "help screen shown\n");
	
	put_byte(sv_out, opcode);
}

void ser_smallkick(int opcode)
{
	put_byte(sv_out, opcode);
}

void ser_bigkick(int opcode)
{
	put_byte(sv_out, opcode);
}

void ser_updateping(int opcode)
{
	int player, ping;
	
	player = get_byte(sv_in);
	ping = get_short(sv_in);
	
	fprintf(stderr, "ping for player %d is %d\n", player, ping);
	
	put_byte(sv_out, opcode);
	put_byte(sv_out, player);
	put_short(sv_out, ping);
}

void ser_updateenter(int opcode)
{
	int player;
	float time;
	
	player = get_byte(sv_in);
	time = get_float(sv_in);
	
	fprintf(stderr, "player %d entered at %f\n", player, time);
	
	put_byte(sv_out, opcode);
	put_byte(sv_out, player);
	put_float(sv_out, time);
}

void ser_updatestatlong(int opcode)
{
	int id;
	long value;
	
	id = get_byte(sv_in);
	value = get_long(sv_in);
	
	fprintf(stderr, "stat: %d = %ld\n", id, value);
//	clprint(2, "statlong: %d = %ld (%08x) diff: %08x\n", id, value, value, value & ~stats[id]);

	stats[id] = value;
		
	put_byte(sv_out, opcode);
	put_byte(sv_out, id);
	put_long(sv_out, value);
}

void ser_muzzleflash(int opcode)
{
	int entity;
	
	entity = get_short(sv_in);

	fprintf(stderr, "muzzleflash on entity %d\n", entity);
	
	put_byte(sv_out, opcode);
	put_short(sv_out, entity);
}

void ser_updateuserinfo(int opcode)
{
	int player;
	long userid;
	char *info, *ptr;
	
	player = get_byte(sv_in);
	userid = get_long(sv_in);
	info = get_string(sv_in);
	
	fprintf(stderr, "user info for player %d [id: %ld]: %s\n", player, userid, info);

	if ((ptr = strstr(info, "name\\"))) {
		/* free() me */
		if (players[player].name)
			free(players[player].name);
		players[player].name = strdup(ptr + 5);
		if ((ptr = strchr(players[player].name, '\\')))
			*ptr = 0;
	}

	if ((ptr = strstr(info, "team\\"))) {
		if (players[player].team)
			free(players[player].team);
		players[player].team = strdup(ptr + 5);
		if ((ptr = strchr(players[player].team, '\\')))
			*ptr = 0;
	}

	put_byte(sv_out, opcode);
	put_byte(sv_out, player);
	put_long(sv_out, userid);
	put_string(sv_out, info);
}

void ser_download(int opcode)
{
	int length, percent, x;
	static int last = -1, totlen = 0;
	char buf[2048];
		
	length = get_short(sv_in);
	percent = get_byte(sv_in);
	if (length == 0xffff && percent == 0)
		fprintf(stderr, "file not found\n");
	else {
		if (percent == 0)
			totlen = 0;
		totlen += length;
		if (last >= 0) {
			int sec;
			
			sec = time(NULL) - last;
			if (!sec)
				sec = 1;
			printf("downloading: %d/%d bytes (%d%%): eta %ds\n", totlen, (int)((float)(length/(percent/100))), percent, (int)(((float)(length/(percent/100))-totlen)/sec));
		}
		last = time(NULL);
		fprintf(stderr, "downloading %d bytes (%d%%)\n", length, percent);
		for (x = 0; x < length; x++)
			buf[x] = get_byte(sv_in);
	}
	
	put_byte(sv_out, opcode);
	put_short(sv_out, length);
	put_byte(sv_out, percent);
	if (!(length == 0xffff && percent == 0))
		for (x = 0; x < length; x++)
			put_byte(sv_out, buf[x]);
}

void ser_playerinfo(int opcode)
{
	int player, unknown, x;
	int mask2, ping;
	long mask;
	float angx, angy, adj, opp;
	float pos[3];
	
	player = get_byte(sv_in);
	players[player].visible = 1;
	mask = get_short(sv_in);

	for (x = 0; x < 3; x++)
		players[player].origin[x] = get_coord(sv_in);
	players[player].frame = get_byte(sv_in);
	
	if (mask & 0x0001) ping = get_byte(sv_in), fprintf(stderr, "ping: %d\n", ping);
	if (mask & 0x0002) {
		mask2 = get_byte(sv_in);
		if (mask2 & 0x01) players[player].angle[0] = get_angle16(sv_in);
		if (mask2 & 0x80) players[player].angle[1] = get_angle16(sv_in);
		if (mask2 & 0x02) players[player].angle[2] = get_angle16(sv_in);
		players[player].going[0] = (mask2 & 0x04) ? get_short(sv_in) : 0;
		players[player].going[1] = (mask2 & 0x08) ? get_short(sv_in) : 0;
		players[player].going[2] = (mask2 & 0x10) ? get_short(sv_in) : 0;
		if (mask2 & 0x20) players[player].flags = get_byte(sv_in);
		if (mask2 & 0x40) players[player].impulse = get_byte(sv_in);
		players[player].load = get_byte(sv_in);
	}
	players[player].speed[0] = (mask & 0x0004) ? get_short(sv_in) : 0;
	players[player].speed[1] = (mask & 0x0008) ? get_short(sv_in) : 0;
	players[player].speed[2] = (mask & 0x0010) ? get_short(sv_in) : 0;
	if (mask & 0x0020) players[player].modelindex = get_byte(sv_in);
	if (mask & 0x0040) unknown = get_byte(sv_in);
	if (mask & 0x0080) players[player].weapon = get_byte(sv_in);
	if (mask & 0x0100) players[player].weaponframe = get_byte(sv_in);

	put_byte(sv_out, opcode);
	put_byte(sv_out, player);
	put_short(sv_out, mask);
	for (x = 0; x < 3; x++)
		put_coord(sv_out, players[player].origin[x]);
	put_byte(sv_out, players[player].frame);
	
	if (mask & 0x0001) put_byte(sv_out, ping);
	if (mask & 0x0002) {
		put_byte(sv_out, mask2);
		if (mask2 & 0x01) put_angle16(sv_out, players[player].angle[0]);
		if (mask2 & 0x80) put_angle16(sv_out, players[player].angle[1]);
		if (mask2 & 0x02) put_angle16(sv_out, players[player].angle[2]);
		if (mask2 & 0x04) put_coord(sv_out, players[player].going[0]);
		if (mask2 & 0x08) put_coord(sv_out, players[player].going[1]);
		if (mask2 & 0x10) put_coord(sv_out, players[player].going[2]);
		if (mask2 & 0x20) put_byte(sv_out, players[player].flags);
		if (mask2 & 0x40) put_byte(sv_out, players[player].impulse);
		put_byte(sv_out, players[player].load);
	}
	if (mask & 0x0004) put_short(sv_out, players[player].speed[0]);
	if (mask & 0x0008) put_short(sv_out, players[player].speed[1]);
	if (mask & 0x0010) put_short(sv_out, players[player].speed[2]);
	if (mask & 0x0020) put_byte(sv_out, players[player].modelindex);
	if (mask & 0x0040) put_byte(sv_out, unknown);
	if (mask & 0x0080) put_byte(sv_out, players[player].weapon);
	if (mask & 0x0100) put_byte(sv_out, players[player].weaponframe);

	/* testy */
	player = find_nearest_player();
	if (player != myid) {
#if 0
	clprint(5, "\n\n\n\nplayer %d:\npos: '%.1f %.1f %.1f'\ndistance: '%.1f %.1f %.1f'\nspeed: '%.1f %.1f %.1f'\ngoing: '%.1f %.1f %.1f'\nangle: '%.1f %.1f %.1f'\nmodel: %d weapon: %d wframe: %d\nflags: %d impulse: %d load: %d",
	player, players[player].origin[0], players[player].origin[1], players[player].origin[2], players[player].origin[0] - players[myid].origin[0], players[player].origin[1] - players[myid].origin[1], players[player].origin[2] - players[myid].origin[2],
	players[player].speed[0], players[player].speed[1], players[player].speed[2],
	players[player].going[0], players[player].going[1], players[player].going[2],
	players[player].angle[0], players[player].angle[1], players[player].angle[2],
	players[player].modelindex, players[player].weapon, players[player].frame, players[player].flags, players[player].impulse, players[player].load);
#endif
	
	if (autoaim) {
		float x1, y1, z1, x2, y2, z2, angx, angy;
		
		x1 = players[myid].origin[0] + (players[myid].cgoing[0]/20.0)*MULT;
		y1 = players[myid].origin[1] + (players[myid].cgoing[1]/20.0)*MULT;
		z1 = players[myid].origin[2] + (players[myid].cgoing[2]/20.0)*MULT;

#if 1
		/* definately not */
		x2 = players[player].origin[0] + (players[player].speed[0])/20.0;
		y2 = players[player].origin[1] + (players[player].speed[1])/20.0;
		z2 = players[player].origin[2] + (players[player].speed[2])/20.0;
#else
		x2 = players[player].origin[0] + (players[player].speed[0]/info.framerate)*MULT;
		y2 = players[player].origin[1] + (players[player].speed[1]/info.framerate)*MULT;
		z2 = players[player].origin[2] + (players[player].speed[2]/info.framerate)*MULT;
#endif
		players[myid].angle[0] = angx = (get_tilt(x1, y1, z1, x2, y2, z2) * 180.0/M_PI);
		players[myid].angle[1] = angy = (get_yaw(x1, y1, x2, y2) * 180.0/M_PI);
		
		if (!players[player].team || strcasecmp(players[myid].team, players[player].team)) {
			put_byte(sv_pending, 0x0a);
			put_angle(sv_pending, angx);
			put_angle(sv_pending, angy);
			put_angle(sv_pending, players[myid].angle[2]);
		}
	}

#if 0
	clprint(5, "\n\n\n\n\n\n\nplayer: %s(%s/%d)\npos: %.1f %.1f %.1f\ndist: %.1f %.1f %.1f\nspeed: %.1f %.1f %.1f\nfps: %.1f\nmyspeed: %.1f %.1f %.1f", 
	players[player].name, players[player].team, player, players[player].origin[0], players[player].origin[1], players[player].origin[2], players[player].origin[0] - players[myid].origin[0], players[player].origin[1] - players[myid].origin[1], players[player].origin[2] - players[myid].origin[2],
	players[player].speed[0], players[player].speed[1], players[player].speed[2],
	info.framerate,
	players[myid].speed[0], players[myid].speed[1], players[myid].speed[2]);
#endif
		clprint(5, "%s\n\n\nplayer: %s(%s)\ndist: %.1f %.1f %.1f", 
		centerprint ? centerprint : "\n\n\n\n\n\n\n\n",
		players[player].name, players[player].team, players[player].origin[0] - players[myid].origin[0], players[player].origin[1] - players[myid].origin[1], players[player].origin[2] - players[myid].origin[2]);
	} else
		clprint(5, "%s\n\npos: %.1f %.1f %.1f", 
		centerprint ? centerprint : "\n\n\n\n\n\n\n\n\n",
		players[player].origin[0], players[player].origin[1], players[player].origin[2]);
}


void ser_nails(int opcode)
{
	/* really ugly, code this right later */
	int nailcount, x, y;
	
	put_byte(sv_out, opcode);
	nailcount = get_byte(sv_in);
	put_byte(sv_out, nailcount);
	for (x = 0; x < nailcount; x++) {
		for (y = 0; y < 5; y++)
			put_byte(sv_out, get_byte(sv_in));
		put_angle(sv_out, get_angle(sv_in));
	}
}

void ser_choke(int opcode)
{
	int packets;
	
	packets = get_byte(sv_in);

	fprintf(stderr, "server hosed: go back %d packets\n", packets);
	
	put_byte(sv_out, opcode);
	put_byte(sv_out, packets);
}

void ser_precachemodels(int opcode)
{
	int first, last, x;
	char *model;
	
	first = get_byte(sv_in);

	if (first == 0) {
		for (x = 1; models[x]; x++)
			free(models[x]);
	}
	

	for (x = first + 1; x < 256; x++) {
		model = get_string(sv_in);
		if (*model)
			models[x] = strdup(model);
		else {
			models[x] = NULL;
			break;
		}
	}

	last = get_byte(sv_in);
	fprintf(stderr, "%d models precached...\n", x);

	put_byte(sv_out, opcode);
	put_byte(sv_out, first);
	for (x = first + 1; models[x]; x++)
		put_string(sv_out, models[x]);
	put_string(sv_out, "");
	put_byte(sv_out, last);
}

void ser_precachesounds(int opcode)
{
	char *sound;
	int first, last, x;
	
	first = get_byte(sv_in);

	if (first == 0) {
		for (x = 1; sounds[x]; x++)
			free(sounds[x]);
	}
	

	for (x = first + 1; x < 256; x++) {
		sound = get_string(sv_in);
		if (*sound)
			sounds[x] = strdup(sound);
		else {
			sounds[x] = NULL;
			break;
		}
	}

	last = get_byte(sv_in);
	fprintf(stderr, "%d sounds precached...\n", x);

	put_byte(sv_out, opcode);
	put_byte(sv_out, first);
	for (x = first + 1; sounds[x]; x++)
		put_string(sv_out, sounds[x]);
	put_string(sv_out, "");
	put_byte(sv_out, last);
}

void ser_entities(int opcode)
{
	unsigned long data, mask, ent;
	
	put_byte(sv_out, opcode);

	while ((data = get_short(sv_in))) {
		ent = data & 0x1ff;
		mask = data & 0xfe00;

		if (mask & 0x8000) mask |= get_byte(sv_in);
		entities[ent].visible = (mask & 0x4000) ? 0 : 1;
		if (mask & 0x0004) entities[ent].modelidx = get_byte(sv_in);
		if (mask & 0x2000) entities[ent].frame = get_byte(sv_in);
		if (mask & 0x0008) entities[ent].colormap = get_byte(sv_in);
		if (mask & 0x0010) entities[ent].skin = get_byte(sv_in);
		if (mask & 0x0020) entities[ent].effects = get_byte(sv_in);
		if (mask & 0x0200) entities[ent].origin[0] = get_coord(sv_in);
		if (mask & 0x0001) entities[ent].angle[0] = get_angle(sv_in);
		if (mask & 0x0400) entities[ent].origin[1] = get_coord(sv_in);
		if (mask & 0x1000) entities[ent].angle[1] = get_angle(sv_in);
		if (mask & 0x0800) entities[ent].origin[2] = get_coord(sv_in);
		if (mask & 0x0002) entities[ent].angle[2] = get_angle(sv_in);
		
		put_short(sv_out, data);
		if (mask & 0x8000) put_byte(sv_out, mask & ~0xfe00);
		if (mask & 0x0004) put_byte(sv_out, entities[ent].modelidx);
		if (mask & 0x2000) put_byte(sv_out, entities[ent].frame);
		if (mask & 0x0008) put_byte(sv_out, entities[ent].colormap);
		if (mask & 0x0010) put_byte(sv_out, entities[ent].skin);
		if (mask & 0x0020) put_byte(sv_out, entities[ent].effects);
		if (mask & 0x0200) put_coord(sv_out, entities[ent].origin[0]);
		if (mask & 0x0001) put_angle(sv_out, entities[ent].angle[0]);
		if (mask & 0x0400) put_coord(sv_out, entities[ent].origin[1]);
		if (mask & 0x1000) put_angle(sv_out, entities[ent].angle[1]);
		if (mask & 0x0800) put_coord(sv_out, entities[ent].origin[2]);
		if (mask & 0x0002) put_angle(sv_out, entities[ent].angle[2]);
	}
	put_short(sv_out, 0);
}

void ser_deltaentities(int opcode)
{
	long data, mask, ent, frame;
	int x;
	
	for (x = 0; x < 32; x++)
		players[x].visible = 0;

	put_byte(sv_out, opcode);
	frame = get_byte(sv_in);
	put_byte(sv_out, frame);

	while ((data = get_short(sv_in))) {
		ent = mask & 0x1ff;
		mask = data & 0xfe00;
		
		if (mask & 0x8000) mask |= get_byte(sv_in);
		fprintf(stderr, "mask: %04lx [got 1]\n", mask);
		entities[ent].visible = (mask & 0x4000) ? 0 : 1;
		if (mask & 0x0004) entities[ent].modelidx = get_byte(sv_in);
		if (mask & 0x2000) entities[ent].frame = get_byte(sv_in);
		if (mask & 0x0008) entities[ent].colormap = get_byte(sv_in);
		if (mask & 0x0010) entities[ent].skin = get_byte(sv_in);
		if (mask & 0x0020) entities[ent].effects = get_byte(sv_in);
		if (mask & 0x0200) entities[ent].origin[0] = get_coord(sv_in);
		if (mask & 0x0001) entities[ent].angle[0] = get_angle(sv_in);
		if (mask & 0x0400) entities[ent].origin[1] = get_coord(sv_in);

		if (mask & 0x1000) entities[ent].angle[1] = get_angle(sv_in);
		if (mask & 0x0800) entities[ent].origin[2] = get_coord(sv_in);
		if (mask & 0x0002) entities[ent].angle[2] = get_angle(sv_in);
		
		put_short(sv_out, data);
		if (mask & 0x8000) put_byte(sv_out, mask & ~0xfe00);
		if (mask & 0x0004) put_byte(sv_out, entities[ent].modelidx);
		if (mask & 0x2000) put_byte(sv_out, entities[ent].frame);
		if (mask & 0x0008) put_byte(sv_out, entities[ent].colormap);
		if (mask & 0x0010) put_byte(sv_out, entities[ent].skin);
		if (mask & 0x0020) put_byte(sv_out, entities[ent].effects);
		if (mask & 0x0200) put_coord(sv_out, entities[ent].origin[0]);
		if (mask & 0x0001) put_angle(sv_out, entities[ent].angle[0]);
		if (mask & 0x0400) put_coord(sv_out, entities[ent].origin[1]);
		if (mask & 0x1000) put_angle(sv_out, entities[ent].angle[1]);
		if (mask & 0x0800) put_coord(sv_out, entities[ent].origin[2]);
		if (mask & 0x0002) put_angle(sv_out, entities[ent].angle[2]);
	}
	put_short(sv_out, 0);
}

void ser_maxspeed(int opcode)
{
	float maxspeed;
	
	maxspeed = get_float(sv_in);
	fprintf(stderr, "max speed: %.1f\n", maxspeed);
	put_byte(sv_out, opcode);
	put_float(sv_out, maxspeed);
}

void ser_entgravity(int opcode)
{
	float grav;
	
	grav = get_float(sv_in);
	fprintf(stderr, "ent gravity: %.1f\n", grav);
	put_byte(sv_out, opcode);
	put_float(sv_out, grav);
}

void ser_setinfo(int opcode)
{
	int player;
	char *key, *data;
	
	player = get_byte(sv_in);
	key = strdup(get_string(sv_in));
	data = get_string(sv_in);
	
	fprintf(stderr, "player %d: %s = %s\n", player, key, data);
	
	if (!strcmp(key, "name")) {
		if (players[player].name)
			free(players[player].name);
		players[player].name = strdup(data);
	}
	if (!strcmp(key, "team")) {
		if (players[player].team)
			free(players[player].team);
		players[player].team = strdup(data);
	}
	
	put_byte(sv_out, opcode);
	put_byte(sv_out, player);
	put_string(sv_out, key);
	put_string(sv_out, data);

	free(key);
}

void ser_serverinfo(int opcode)
{
	char *key, *data;
	
	key = get_string(sv_in);
	data = get_string(sv_in);
	
	fprintf(stderr, "server: %s = %s\n", key, data);
	
	put_byte(sv_out, opcode);
	put_string(sv_out, key);
	put_string(sv_out, data);
}

void ser_updatepl(int opcode)
{
	int player, loss;
	
	player = get_byte(sv_in);
	loss = get_byte(sv_in);
	
	fprintf(stderr, "player %d has %d%% packet loss\n", player, loss);
	
	put_byte(sv_out, opcode);
	put_byte(sv_out, player);
	put_byte(sv_out, loss);
}
